#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "Common.h"
#include "OpCodes.h"
#include "Class.h"
#include "Heap.h"
#include "Stack.h"
#include "JVM.h"

//-----------------------------------------------------------------------------
// Name: cHeap::cHeap()
//-----------------------------------------------------------------------------
cHeap::cHeap() 
{
	m_pMainMemory = NULL;
	m_iMemorySize = 0;
	m_iTop = 0;
	m_iBoundry = 0;
	m_pOwner = NULL;
}

//-----------------------------------------------------------------------------
// Name: cHeap::~cHeap()
//-----------------------------------------------------------------------------
cHeap::~cHeap()
{
	SAFE_DELETE_ARRAY(m_pMainMemory);
}

//-----------------------------------------------------------------------------
// Name: cHeap::SetOwner()
//-----------------------------------------------------------------------------
void cHeap::SetOwner(cJVM *pOwner)
{
	m_pOwner = pOwner;
}

//-----------------------------------------------------------------------------
// Name: cHeap::GetNumAllocatedBytes()
//-----------------------------------------------------------------------------
u4 cHeap::GetNumAllocatedBytes() 
{
	 if(m_iTop < (m_iMemorySize / 2))
		 return m_iTop;
	 else
		 return m_iTop - m_iMemorySize / 2;
}

//-----------------------------------------------------------------------------
// Name: cHeap::Init()
//-----------------------------------------------------------------------------
void cHeap::Init(u4 iHeapSize)
{
	SAFE_DELETE_ARRAY(m_pMainMemory);
	m_pMainMemory = new u1[iHeapSize];
	memset(m_pMainMemory,0,iHeapSize);
	m_iMemorySize = iHeapSize;
	m_iTop = 0;
	m_iBoundry = iHeapSize / 2;
}

//-----------------------------------------------------------------------------
// Name: cHeap::SwapHeapspace()
//-----------------------------------------------------------------------------
void cHeap::SwapHeapspace()
{
	//swap spaces during garbage collection
	if(m_iTop < (m_iMemorySize / 2))
	{
		m_iTop = m_iBoundry;
		m_iBoundry = m_iMemorySize;
	}
	else
	{
		m_iTop = 0;
		m_iBoundry = m_iMemorySize / 2;
	}
}

//-----------------------------------------------------------------------------
// Name: cHeap::AllocateClass()
//-----------------------------------------------------------------------------
OBJECT* cHeap::AllocateClass(cJavaClass *pClass)
{
	//get size + header
	u4 iSize = pClass->GetSize();
	iSize += sizeof(OBJECT);

	if(m_iTop + iSize >= m_iBoundry) //no more mem -> run GC
	{
		m_pOwner->RunGC();

		if(m_iTop + iSize >= m_iBoundry)
			return NULL; //out of mem, even after GC run
	}

	//set memory
	OBJECT *pObj = (OBJECT*)&m_pMainMemory[m_iTop];
	//zero memory
	memset(pObj,0,iSize);
	m_iTop += iSize;

	//fill header
	pObj->iArraySize = 1;
	pObj->pClass = pClass;
	pObj->iSize = iSize;

	return pObj;
}

//-----------------------------------------------------------------------------
// Name: cHeap::AllocateArray()
//-----------------------------------------------------------------------------
OBJECT* cHeap::AllocateArray(cJavaClass *pClass, int iCount, int iItemSize)
{
	//compute size
	u4 iSize = iCount * iItemSize;
	iSize += sizeof(OBJECT);

	if(m_iTop + iSize >= m_iBoundry) //no more mem -> run GC
	{
		m_pOwner->RunGC();

		if(m_iTop + iSize >= m_iBoundry)
			return NULL; //out of mem, even after GC run
	}

	//set memory
	OBJECT *pObj = (OBJECT*)&m_pMainMemory[m_iTop];
	//zero memory
	memset(pObj,0,iSize);
	m_iTop += iSize;

	//update header
	pObj->iArraySize = iCount;
	pObj->pClass = pClass;
	pObj->iSize = iSize;

	return pObj;
}

//-----------------------------------------------------------------------------
// Name: cHeap::RelocateObject()
//-----------------------------------------------------------------------------
OBJECT* cHeap::RelocateObject(OBJECT *pObj)
{
	OBJECT *pNewObj = (OBJECT*)&m_pMainMemory[m_iTop];
	memcpy(pNewObj,pObj,pObj->iSize);
	m_iTop += pObj->iSize;


	//make note of new address
	pObj->iSize = (u4)pNewObj;
	//flag realocated object
	pObj->iArraySize = 0;

	return pNewObj;
}