#define _CRT_SECURE_NO_WARNINGS

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <map>
#include <string>

#include "Common.h"
#include "OpCodes.h"
#include "Class.h"
#include "Heap.h"
#include "Stack.h"
#include "JVM.h"

//----------------------------------------------------------------------------
// TESTING ONLY
//----------------------------------------------------------------------------
int main()
{
	cJVM	jvm;
	jvm.Initialize(128,700);
	jvm.LoadMain("Main.class");

	jvm.Run();
}

//-----------------------------------------------------------------------------
// Name: cJVM::cJVM()
//-----------------------------------------------------------------------------
cJVM::cJVM()
{
	m_pPC = NULL;
	m_bError = false;
	m_Heap.SetOwner(this);

	for(int i = 0;i < INST_NUMINST;++i)
		Instructions[i] = NULL;

	//assign instruction functions
	Instructions[INST_NEW] = &cJVM::Inst_New;
	Instructions[INST_DUP] = &cJVM::Inst_Dup;
	Instructions[INST_INVOKESPECIAL] = &cJVM::Inst_InvokeSpecial;
	Instructions[INST_INVOKEVIRTUAL] = &cJVM::Inst_InvokeVirtual;
	Instructions[INST_DLOAD] = &cJVM::Inst_DLoad;
	Instructions[INST_DLOAD_1] = &cJVM::Inst_DLoad_1;
	Instructions[INST_ALOAD] = &cJVM::Inst_ALoad;
	Instructions[INST_ALOAD_0] = &cJVM::Inst_ALoad_0;
	Instructions[INST_ALOAD_1] = &cJVM::Inst_ALoad_1;
	Instructions[INST_ALOAD_2] = &cJVM::Inst_ALoad_2;
	Instructions[INST_ALOAD_3] = &cJVM::Inst_ALoad_3;
	Instructions[INST_RETURN] = &cJVM::Inst_Return;
	Instructions[INST_IRETURN] = &cJVM::Inst_IReturn;
	Instructions[INST_ARETURN] = &cJVM::Inst_AReturn;
	Instructions[INST_ASTORE_1] = &cJVM::Inst_AStore_1;
	Instructions[INST_ASTORE_2] = &cJVM::Inst_AStore_2;
	Instructions[INST_ASTORE_3] = &cJVM::Inst_AStore_3;
	Instructions[INST_ISTORE] = &cJVM::Inst_IStore;
	Instructions[INST_ISTORE_1] = &cJVM::Inst_IStore_1;
	Instructions[INST_ISTORE_0] = &cJVM::Inst_IStore_0;
	Instructions[INST_ISTORE_2] = &cJVM::Inst_IStore_2;
	Instructions[INST_ISTORE_3] = &cJVM::Inst_IStore_3;
	Instructions[INST_ICONST_0] = &cJVM::Inst_IConst_0;
	Instructions[INST_ICONST_1] = &cJVM::Inst_IConst_1;
	Instructions[INST_ICONST_2] = &cJVM::Inst_IConst_2;
	Instructions[INST_ICONST_3] = &cJVM::Inst_IConst_3;
	Instructions[INST_ICONST_4] = &cJVM::Inst_IConst_4;
	Instructions[INST_ICONST_5] = &cJVM::Inst_IConst_5;
	Instructions[INST_ILOAD] = &cJVM::Inst_ILoad;
	Instructions[INST_ILOAD_0] = &cJVM::Inst_ILoad_0;
	Instructions[INST_ILOAD_1] = &cJVM::Inst_ILoad_1;
	Instructions[INST_ILOAD_2] = &cJVM::Inst_ILoad_2;
	Instructions[INST_ILOAD_3] = &cJVM::Inst_ILoad_3;
	Instructions[INST_IADD] = &cJVM::Inst_IAdd;
	Instructions[INST_POP] = &cJVM::Inst_Pop;
	Instructions[INST_IASTORE] = &cJVM::Inst_IAStore;
	Instructions[INST_BIPUSH] = &cJVM::Inst_BIPush;
	Instructions[INST_PUTSTATIC] = &cJVM::Inst_PutStatic;
	Instructions[INST_GETSTATIC] = &cJVM::Inst_GetStatic;
	Instructions[INST_PUTFIELD] = &cJVM::Inst_PutField;
	Instructions[INST_GETFIELD] = &cJVM::Inst_GetField;
	Instructions[INST_NEWARRAY] = &cJVM::Inst_NewArray;
	Instructions[INST_GOTO] = &cJVM::Inst_GoTo;
	Instructions[INST_BALOAD] = &cJVM::Inst_BALoad;
	Instructions[INST_IINC] = &cJVM::Inst_IInc;
	Instructions[INST_ARRAYLENGTH] = &cJVM::Inst_ArrayLength;
	Instructions[INST_IF_ICMPLT] = &cJVM::Inst_If_ICMPLT;
	Instructions[INST_SIPUSH] = &cJVM::Inst_SIPush;
	Instructions[INST_BASTORE] = &cJVM::Inst_BAStore;
	Instructions[INST_IALOAD] = &cJVM::Inst_IALoad;
	Instructions[INST_AALOAD] = &cJVM::Inst_AALoad;
	Instructions[INST_LDC2_W] = &cJVM::Inst_LDC2_W;
	Instructions[INST_DADD] = &cJVM::Inst_DAdd;
	Instructions[INST_DRETURN] = &cJVM::Inst_DReturn;
	Instructions[INST_D2I] = &cJVM::Inst_D2I;
	Instructions[INST_ANEWARRAY] = &cJVM::Inst_ANewArray;
	Instructions[INST_DCONST_0] = &cJVM::Inst_DConst_0;
	Instructions[INST_DCONST_1] = &cJVM::Inst_DConst_1;	
	Instructions[INST_AASTORE] = &cJVM::Inst_AAStore;	
	Instructions[INST_INVOKESTATIC] = &cJVM::Inst_InvokeStatic;
	Instructions[INST_DSTORE_1] = &cJVM::Inst_DStore_1;
	Instructions[INST_ACONST_NULL] = &cJVM::Inst_AConst_Null;
}

//-----------------------------------------------------------------------------
// Name: cJVM::~cJVM()
//-----------------------------------------------------------------------------
cJVM::~cJVM()
{
	//make sure to release allocated memory
	FreeAllClasses();
}

//-----------------------------------------------------------------------------
// Name: cJVM::Error()
//-----------------------------------------------------------------------------
void cJVM::Error(char *strError)
{
	//stop execution
	m_pPC = NULL;
	//print error
	printf("%s\n",strError);
	//clear stack
	m_Stack.Clear();
	//raise flag
	m_bError = true;
}

//-----------------------------------------------------------------------------
// Name: cJVM::FreeAllClasses()
//-----------------------------------------------------------------------------
void cJVM::FreeAllClasses()
{
	//get rid of them, one by one...
	for(ClassMapItor itor = m_Classes.begin();itor != m_Classes.end();++itor)
	{
		SAFE_DELETE(itor->second);
	}
	m_Classes.clear();
}

//-----------------------------------------------------------------------------
// Name: cJVM::Initialize()
//-----------------------------------------------------------------------------
bool cJVM::Initialize(u4 iStackSize, u4 iHeapSize)
{
	//one time initialization
	//stack size - in stack values
	//heap size - in bytes
	m_Stack.Resize(iStackSize);
	m_Heap.Init(iHeapSize);
	return true;
}

//-----------------------------------------------------------------------------
// Name: cJVM::LoadMain()
//-----------------------------------------------------------------------------
bool cJVM::LoadMain(const char *strFilename)
{
	//create class instance
	cJavaClass *pClass = new cJavaClass(this);

	try
	{
		//try to load class from file
		pClass->LoadFromFile(strFilename);

		//add class to map
		std::pair<ClassMapItor,bool> pr = m_Classes.insert(ClassMapValue(std::string(strFilename),pClass));
		if(pr.second == false) //unable to insert
		{
			throw std::string("Internal Error");
		}

		//try to resolve constant pool references
		pClass->ResolveSymbolicReferences();
		//init static(constant) data 
		pClass->InitializeStaticData();

		//run classe's initialization method (clinit)
		if(!this->InitializeClass(pClass))
		{
			throw std::string("Class's initialization method failed.");
		}
	}
	catch(std::string strError)
	{
		m_Classes.erase(std::string(strFilename));
		SAFE_DELETE(pClass);
		printf("Unable to load class \"%s\" - %s\n",strFilename,strError.c_str());
		return false;
	}

	//check for main
	if(pClass->m_pMainMethod == NULL)
		return false;

	//prepare main method for running
	//method's frame
	m_Stack.PushMethodFrame(pClass->m_pMainMethod, NULL, 0);
	//program counter
	m_pPC = pClass->m_pMainMethod->pCode;
	//and currnet class
	m_pCurClass = pClass;

	return true;
}

//-----------------------------------------------------------------------------
// Name: cJVM::LoadClass()
//-----------------------------------------------------------------------------
cJavaClass* cJVM::LoadClass(const char *strFilename)
{
	//load class from java class file
	//class names in constant pool don't have extension, so add one
	std::string s = strFilename;
	s += ".class";

	//check if this class is already loaded
	ClassMapItor i = m_Classes.find(s);
	if(i != m_Classes.end())
	{
		return i->second;
	}

	//load class
	cJavaClass *pClass = new cJavaClass(this);

	try
	{
		pClass->LoadFromFile(s.c_str());

		//add class to map
		std::pair<ClassMapItor,bool> pr = m_Classes.insert(ClassMapValue(s,pClass));
		if(pr.second == false) //unable to insert
		{
			//this really shouldn't happen, so just for sure
			throw std::string("Internal Error");
		}

		//try to resolve constant pool references
		pClass->ResolveSymbolicReferences();
		//init static(constant) data 
		pClass->InitializeStaticData();

		//run classe's initialization method (clinit)
		if(!this->InitializeClass(pClass))
		{
			throw std::string("Class's initialization method failed.");
		}
	}
	catch(std::string strError)
	{
		m_Classes.erase(s);
		SAFE_DELETE(pClass);
		printf("Unable to load class \"%s\" - %s\n",strFilename,strError.c_str());
		return NULL;
	}	

	return pClass;
}

//-----------------------------------------------------------------------------
// Name: cJVM::Run()
//-----------------------------------------------------------------------------
bool cJVM::Run()
{
	//start execution prepared metohd
	//first, check if any method was prepared
	if(m_pPC == NULL)
		return false;

	//now, go trought byto code, one instruction at a time
	u1 iOpCode;
	while(m_pPC != NULL)
	{
		//fetch opcode
		iOpCode = *m_pPC;
		//printf("%d\n",iOpCode); //DEBUG
		m_pPC++;

		//this section could be erased after all instrucionts are implemented
		if(this->Instructions[iOpCode] == NULL)
		{ //opcode isn't implemented
			printf("Unknown opcode: %d\n",iOpCode);
			return false;
		}

		//call appropriate function based on opcode
		(this->*Instructions[iOpCode])();
	}

	if(m_bError)
	{
		m_bError = false;
		return false;
	}

	//DEBUG OUTPUT - get rid of when no longer needed
	printf("\nAnd the result is: %d\n",m_Stack.GetOperand(3).iInt);

	return true;
}


//-----------------------------------------------------------------------------
// Name: cJVM::InitializeClass()
//-----------------------------------------------------------------------------
bool cJVM::InitializeClass(cJavaClass *pClass)
{
	//call classe's initialization method (clinit)
	if(pClass->m_pClassInit == NULL)
		return true;

	//setup stak frame
	m_Stack.PushMethodFrame(pClass->m_pClassInit, m_pPC, 0);
	m_pPC = pClass->m_pClassInit->pCode;
	m_pCurClass = pClass;

	//execute bytecode
	u1 iOpCode;
	while(m_pPC != NULL)
	{
		iOpCode = *m_pPC;
		//printf("%d\n",iOpCode);
		m_pPC++;

		if(this->Instructions[iOpCode] == NULL)
		{ //opcode isn't implemented
			printf("Unknown opcode: %d\n",iOpCode);
			return false;
		}
		(this->*Instructions[iOpCode])();
	}

	if(m_bError)
	{
		m_bError = false;
		return false;
	}

	return true;
}

//-----------------------------------------------------------------------------
// Name: cJVM::ReadU2()
//-----------------------------------------------------------------------------
u2 cJVM::ReadU2() 
{
	//read two byte value from pc
	u1 i1 = *m_pPC;
	++m_pPC;
	u1 i2 = *m_pPC;
	++m_pPC;

	//it's big endian and we need little

	return (u2)((i1 << 8) | i2);
}

//-----------------------------------------------------------------------------
// Name: cJVM::GetDouble()
//-----------------------------------------------------------------------------
void cJVM::GetDouble(double &d,u4 iHighBytes,u4 iLowBytes)
{
	//compute double from high and low bytes
	//just for clarification, long long in C++ is the same as long in java (8-byte)
	long long l = (long long)(((long long)iHighBytes << 32) | iLowBytes);
	memcpy(&d,&l,sizeof(double));
}

//-----------------------------------------------------------------------------
// Name: cJVM::GetLong()
//-----------------------------------------------------------------------------
void cJVM::GetLong(long long &l,u4 iHighBytes,u4 iLowBytes)
{
	//compute long from high and low bytes
	//just for clarification, long long in C++ is the same as long in java (8-byte)
	l = (long long)(((long long)iHighBytes << 32) | iLowBytes);
}

//-----------------------------------------------------------------------------
// Name: cJVM::GetBytesFromDouble()
//-----------------------------------------------------------------------------
void cJVM::GetBytesFromDouble(double &d,u4 &iHighBytes,u4 &iLowBytes)
{
	//split double to high and low bytes
	u1 *pData = (u1*)&d;

	memcpy(&iLowBytes,pData,sizeof(u4));
	pData += sizeof(u4);
	memcpy(&iHighBytes,pData,sizeof(u4));		
}

//-----------------------------------------------------------------------------
// Name: cJVM::GetBytesFromLong()
//-----------------------------------------------------------------------------
void cJVM::GetBytesFromLong(long long &l,u4 &iHighBytes,u4 &iLowBytes)
{
	//split long to high and low bytes
	u1 *pData = (u1*)&l;

	memcpy(&iLowBytes,pData,sizeof(u4));
	pData += sizeof(u4);
	memcpy(&iHighBytes,pData,sizeof(u4));		
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_New()
//-----------------------------------------------------------------------------
void cJVM::Inst_New()
{
	//constant pool index to class we are about to create
	u2 iIndex = ReadU2();

	//get class pointer
	cJavaClass *pClass = m_pCurClass->GetClassFromPool(iIndex);
	//allocate memory
	OBJECT *pObj = m_Heap.AllocateClass(pClass);

	//check for out-of-mem
	if(pObj == NULL)
	{
		Error("Error: Out-of-memory.");
		return;
	}

	//push result on top of operand stack
	m_Stack.PushOperandObject(pObj);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_NewArray()
//-----------------------------------------------------------------------------
void cJVM::Inst_NewArray()
{
	//first, get data type
	u1 iType = *m_pPC;
	++m_pPC;

	//get item size
	u2 iItemSize = 0;
	switch(iType)
	{
	case 4: //char
	case 5: //byte
	case 8: //bool
		iItemSize = sizeof(char);
		break;
	case 9: //short
		iItemSize = sizeof(short);
		break;
	case 10: //int
		iItemSize = sizeof(int);
		break;
	case 11: //long
		iItemSize = sizeof(long long);
		break;
	case 6: //float
		iItemSize = sizeof(float);
		break;
	case 7: //double
		iItemSize = sizeof(double);
		break;
	};

	//get array size
	int iCount = m_Stack.PopOperand().iInt;

	//allocate memory
	OBJECT *pObj = m_Heap.AllocateArray(NULL,iCount,iItemSize);

	//check for out-of-mem
	if(pObj == NULL)
	{
		Error("Error: Out-of-memory.");
		return;
	}

	//push result on top of operand stack
	m_Stack.PushOperandObject(pObj);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_ANewArray()
//-----------------------------------------------------------------------------
void cJVM::Inst_ANewArray()
{
	//get constant pool class index
	u2 iIndex = ReadU2();
	//get class info
	cJavaClass *pClass = m_pCurClass->GetClassFromPool(iIndex);

	//get item size
	u2 iItemSize = sizeof(OBJECT*);
	
	//get array size
	int iCount = m_Stack.PopOperand().iInt;

	//allocate memory
	OBJECT *pObj = m_Heap.AllocateArray(pClass,iCount,iItemSize);

	//check for out-of-mem
	if(pObj == NULL)
	{
		Error("Error: Out-of-memory.");
		return;
	}

	//push result on top of operand stack
	m_Stack.PushOperandObject(pObj);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_Dup()
//-----------------------------------------------------------------------------
void cJVM::Inst_Dup()
{
	//what else to say
	m_Stack.DuplicateTopOperand();
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_InvokeVirtual()
//-----------------------------------------------------------------------------
void cJVM::Inst_InvokeVirtual()
{
	//get constant pool method index
	u2 iIndex = ReadU2();
	//get method info
	METHOD_INFO *pMethod = m_pCurClass->GetMethodFromPool(iIndex);
	//add one argument for "this" reference
	u2 iNumArgs = pMethod->iNumArguments + 1;

	//push frame
	m_Stack.PushMethodFrame(pMethod,m_pPC,iNumArgs);

	//update cur class and pc
	m_pCurClass = pMethod->pOwner;
	m_pPC = pMethod->pCode;
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_InvokeStatic()
//-----------------------------------------------------------------------------
void cJVM::Inst_InvokeStatic()
{
	//get constant pool method index
	u2 iIndex = ReadU2();
	//get method info
	METHOD_INFO *pMethod = m_pCurClass->GetMethodFromPool(iIndex);
	//get num arguments
	u2 iNumArgs = pMethod->iNumArguments;
	//push frame
	m_Stack.PushMethodFrame(pMethod,m_pPC,iNumArgs);

	//update cur class and pc
	m_pCurClass = pMethod->pOwner;
	m_pPC = pMethod->pCode;
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_InvokeSpecial()
//-----------------------------------------------------------------------------
void cJVM::Inst_InvokeSpecial()
{
	//get constant pool method index
	u2 iIndex = ReadU2();
	//get method info
	METHOD_INFO *pMethod = m_pCurClass->GetMethodFromPool(iIndex);
	//add one argument for "this" reference
	u2 iNumArgs = pMethod->iNumArguments + 1;
	//push frame
	m_Stack.PushMethodFrame(pMethod,m_pPC,iNumArgs);

	//update cur class and pc
	m_pCurClass = pMethod->pOwner;
	m_pPC = pMethod->pCode;
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_ALoad()
//-----------------------------------------------------------------------------
void cJVM::Inst_ALoad()
{
	m_Stack.PushVariable(*m_pPC);
	++m_pPC;
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_DLoad()
//-----------------------------------------------------------------------------
void cJVM::Inst_DLoad()
{
	m_Stack.PushVariableDouble(*m_pPC);
	++m_pPC;
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_DLoad_1()
//-----------------------------------------------------------------------------
void cJVM::Inst_DLoad_1()
{
	m_Stack.PushVariableDouble(1);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_ALoad_0()
//-----------------------------------------------------------------------------
void cJVM::Inst_ALoad_0()
{
	m_Stack.PushVariable(0);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_ALoad_1()
//-----------------------------------------------------------------------------
void cJVM::Inst_ALoad_1()
{
	m_Stack.PushVariable(1);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_ALoad_2()
//-----------------------------------------------------------------------------
void cJVM::Inst_ALoad_2()
{
	m_Stack.PushVariable(2);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_ALoad_3()
//-----------------------------------------------------------------------------
void cJVM::Inst_ALoad_3()
{
	m_Stack.PushVariable(3);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_Return()
//-----------------------------------------------------------------------------
void cJVM::Inst_Return()
{
	//return from method
	m_pPC = m_Stack.PopMethodFrame(&m_pCurClass);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_IReturn()
//-----------------------------------------------------------------------------
void cJVM::Inst_IReturn()
{
	//return and push result to operand stack
	int ret = m_Stack.PopOperand().iInt;
	m_pPC = m_Stack.PopMethodFrame(&m_pCurClass);
	m_Stack.PushOperandInt(ret);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_AReturn()
//-----------------------------------------------------------------------------
void cJVM::Inst_AReturn()
{
	//return and push result to operand stack
	OBJECT *pRet = m_Stack.PopOperand().pObject;
	m_pPC = m_Stack.PopMethodFrame(&m_pCurClass);
	m_Stack.PushOperandObject(pRet);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_DReturn()
//-----------------------------------------------------------------------------
void cJVM::Inst_DReturn()
{
	//return and push double to operand stack
	int iHighBytes,iLowBytes;
	iLowBytes = m_Stack.PopOperand().iInt;
	iHighBytes = m_Stack.PopOperand().iInt;

	m_pPC = m_Stack.PopMethodFrame(&m_pCurClass);

	m_Stack.PushOperandInt(iHighBytes);
	m_Stack.PushOperandInt(iLowBytes);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_AStore_1()
//-----------------------------------------------------------------------------
void cJVM::Inst_AStore_1()
{
	m_Stack.SetVariableObject(m_Stack.PopOperand().pObject, 1);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_AStore_2()
//-----------------------------------------------------------------------------
void cJVM::Inst_AStore_2()
{
	m_Stack.SetVariableObject(m_Stack.PopOperand().pObject, 2);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_AStore_3()
//-----------------------------------------------------------------------------
void cJVM::Inst_AStore_3()
{
	m_Stack.SetVariableObject(m_Stack.PopOperand().pObject, 3);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_IStore_0()
//-----------------------------------------------------------------------------
void cJVM::Inst_IStore_0()
{
	m_Stack.SetVariableInt(m_Stack.PopOperand().iInt, 0);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_DStore_1()
//-----------------------------------------------------------------------------
void cJVM::Inst_DStore_1()
{
	//double -> store two items
	m_Stack.SetVariableInt(m_Stack.PopOperand().iInt, 2);
	m_Stack.SetVariableInt(m_Stack.PopOperand().iInt, 1);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_IStore()
//-----------------------------------------------------------------------------
void cJVM::Inst_IStore()
{
	u1 iIndex = *m_pPC;
	m_pPC++;

	m_Stack.SetVariableInt(m_Stack.PopOperand().iInt, iIndex);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_IStore_1()
//-----------------------------------------------------------------------------
void cJVM::Inst_IStore_1()
{
	m_Stack.SetVariableInt(m_Stack.PopOperand().iInt, 1);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_IStore_2()
//-----------------------------------------------------------------------------
void cJVM::Inst_IStore_2()
{
	m_Stack.SetVariableInt(m_Stack.PopOperand().iInt, 2);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_IStore_3()
//-----------------------------------------------------------------------------
void cJVM::Inst_IStore_3()
{
	m_Stack.SetVariableInt(m_Stack.PopOperand().iInt, 3);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_IConst_0()
//-----------------------------------------------------------------------------
void cJVM::Inst_IConst_0()
{
	m_Stack.PushOperandInt(0);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_IConst_1()
//-----------------------------------------------------------------------------
void cJVM::Inst_IConst_1()
{
	m_Stack.PushOperandInt(1);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_IConst_2()
//-----------------------------------------------------------------------------
void cJVM::Inst_IConst_2()
{
	m_Stack.PushOperandInt(2);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_IConst_3()
//-----------------------------------------------------------------------------
void cJVM::Inst_IConst_3()
{
	m_Stack.PushOperandInt(3);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_IConst_4()
//-----------------------------------------------------------------------------
void cJVM::Inst_IConst_4()
{
	m_Stack.PushOperandInt(4);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_IConst_5()
//-----------------------------------------------------------------------------
void cJVM::Inst_IConst_5()
{
	m_Stack.PushOperandInt(5);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_ILoad()
//-----------------------------------------------------------------------------
void cJVM::Inst_ILoad()
{
	u1 iIndex = *m_pPC;
	m_pPC++;

	m_Stack.PushVariable(iIndex);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_ILoad_0()
//-----------------------------------------------------------------------------
void cJVM::Inst_ILoad_0()
{
	m_Stack.PushVariable(0);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_ILoad_1()
//-----------------------------------------------------------------------------
void cJVM::Inst_ILoad_1()
{
	m_Stack.PushVariable(1);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_ILoad_2()
//-----------------------------------------------------------------------------
void cJVM::Inst_ILoad_2()
{
	m_Stack.PushVariable(2);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_ILoad_3()
//-----------------------------------------------------------------------------
void cJVM::Inst_ILoad_3()
{
	m_Stack.PushVariable(3);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_AConst_Null()
//-----------------------------------------------------------------------------
void cJVM::Inst_AConst_Null()
{
	m_Stack.PushOperandObject(NULL);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_IAdd()
//-----------------------------------------------------------------------------
void cJVM::Inst_IAdd()
{
	m_Stack.PushOperandInt(m_Stack.PopOperand().iInt + m_Stack.PopOperand().iInt);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_Pop()
//-----------------------------------------------------------------------------
void cJVM::Inst_Pop()
{
	m_Stack.PopOperand();
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_IAStore()
//-----------------------------------------------------------------------------
void cJVM::Inst_IAStore()
{
	//get value
	int iValue = m_Stack.PopOperand().iInt;
	//get array index
	int iIndex = m_Stack.PopOperand().iInt;
	//get array
	OBJECT *pObj = m_Stack.PopOperand().pObject;

	//check for null reference
	if(pObj == NULL)
	{
		Error("Error: Null reference.");
		return;
	}
	//check array index
	if(iIndex >= (int)pObj->iArraySize)
	{
		Error("Error: Array out of bounds.");
		return;
	}

	//get size of one item in array
	int iItemSize = (pObj->iSize - sizeof(OBJECT)) / pObj->iArraySize;
	//skip object header
	u1 *pArray = (u1*)(pObj + 1);
	//move to exact position 
	pArray = pArray + (iIndex * iItemSize);
	//copy value
	memcpy(pArray,&iValue,sizeof(u4));
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_BIPush()
//-----------------------------------------------------------------------------
void cJVM::Inst_BIPush()
{
	int iValue = (int)(*m_pPC);
	++m_pPC;
	m_Stack.PushOperandInt(iValue);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_PutStatic()
//-----------------------------------------------------------------------------
void cJVM::Inst_PutStatic()
{
	//write static
	//get constant pool index of field
	u2 iIndex = ReadU2();
	//get field info
	FIELD_INFO *pField = m_pCurClass->GetFieldFromPool(iIndex);

	//determine field type
	if(pField->iSize == sizeof(double))
	{
		int i = m_Stack.PopOperand().iInt;
		//low bytes
		pField->pOwner->SetStaticValue(pField->iOffset + sizeof(double)/2,&i,sizeof(double)/2);
		i = m_Stack.PopOperand().iInt;
		//high bytes
		pField->pOwner->SetStaticValue(pField->iOffset,&i,sizeof(double)/2);
		
	}
	else if(pField->iSize == sizeof(short)) //short
	{
		short s = (short)m_Stack.PopOperand().iInt;			
		pField->pOwner->SetStaticValue(pField->iOffset,&s,pField->iSize);
	}
	else if(pField->iSize == sizeof(char)) //char,bool, byte
	{
		char c = (char)m_Stack.PopOperand().iInt;
		pField->pOwner->SetStaticValue(pField->iOffset,&c,pField->iSize);
	}
	else //everything else is 4-bytes
	{
		int i = m_Stack.PopOperand().iInt;
		pField->pOwner->SetStaticValue(pField->iOffset,&i,pField->iSize);
	}
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_GetStatic()
//-----------------------------------------------------------------------------
void cJVM::Inst_GetStatic()
{
	//load static
	//get constant pool index of field
	u2 iIndex = ReadU2();
	//get field info
	FIELD_INFO *pField = m_pCurClass->GetFieldFromPool(iIndex);
	//determine field size
	if(pField->iSize == sizeof(double))
	{
		int i;			
		//high bytes
		pField->pOwner->GetStaticValue(pField->iOffset,&i,sizeof(double)/2);
		m_Stack.PushOperandInt(i);
		//low bytes
		pField->pOwner->GetStaticValue(pField->iOffset + sizeof(double)/2,&i,sizeof(double)/2);
		m_Stack.PushOperandInt(i);
	}
	else if(pField->iSize == sizeof(short)) //short
	{
		short s;
		pField->pOwner->GetStaticValue(pField->iOffset,&s,pField->iSize);
		m_Stack.PushOperandInt((int)s);
	}
	else if(pField->iSize == sizeof(char)) //char, bool, byte
	{
		char c;
		pField->pOwner->GetStaticValue(pField->iOffset,&c,pField->iSize);
		m_Stack.PushOperandInt((int)c);
	}
	else //everything else is four byte
	{
		STACK_ITEM i;
		pField->pOwner->GetStaticValue(pField->iOffset,&i,pField->iSize);
		m_Stack.PushOperand(i,pField->bReference); 
	}

}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_PutField()
//-----------------------------------------------------------------------------
void cJVM::Inst_PutField()
{
	//get field info from constant pool
	u2 iIndex = ReadU2();
	FIELD_INFO *pField = m_pCurClass->GetFieldFromPool(iIndex);

	//get value to pot
	int item1 = m_Stack.PopOperand().iInt;
	int item2;
	//if field is double or long, fetch second part of value
	if(pField->iSize == sizeof(double))
		item2 = m_Stack.PopOperand().iInt;

	//now, get object
	OBJECT *pObj = m_Stack.PopOperand().pObject;

	//check for null reference
	if(pObj == NULL)
	{
		Error("Error: Null reference.");
		return;
	}

	//skip object header
	u1 *pData = (u1*)(pObj + 1);

	//based on field data type copy value
	if(pField->iSize == sizeof(double))
	{ //double, long
		//low bytes
		memcpy(&pData[pField->iOffset + sizeof(double)/2], &item1, sizeof(double)/2);
		//high bytes
		memcpy(&pData[pField->iOffset], &item2, sizeof(double)/2);
	}
	else if(pField->iSize == sizeof(short))
	{
		short s = (short)item1;
		memcpy(&pData[pField->iOffset], &s, pField->iSize);
	}
	else if(pField->iSize == sizeof(char))
	{
		char c = (char)item1;
		memcpy(&pData[pField->iOffset], &c, pField->iSize);
	}
	else //everything else if four-byte
	{
		memcpy(&pData[pField->iOffset], &item1, pField->iSize);
	}
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_GetField()
//-----------------------------------------------------------------------------
void cJVM::Inst_GetField()
{
	//get field info
	u2 iIndex = ReadU2();
	FIELD_INFO *pField = m_pCurClass->GetFieldFromPool(iIndex);
	//get object
	OBJECT *pObj = m_Stack.PopOperand().pObject;
	//check for null reference
	if(pObj == NULL)
	{
		Error("Error: Null reference.");
		return;
	}

	//skip object header
	u1 *pData = (u1*)(pObj + 1);

	STACK_ITEM item;
	//based on field data type fetch value
	if(pField->iSize == sizeof(double))
	{			
		//high bytes
		memcpy(&item, &pData[pField->iOffset], sizeof(double)/2);
		m_Stack.PushOperand(item,0);

		//low bytes
		memcpy(&item, &pData[pField->iOffset + sizeof(double)/2], sizeof(double)/2);
		m_Stack.PushOperand(item,0);
	}
	else if(pField->iSize == sizeof(short))
	{
		short s = (short)pData[pField->iOffset];
		m_Stack.PushOperandInt((int)s);
	}
	else if(pField->iSize == sizeof(char))
	{
		char c = (char)pData[pField->iOffset];
		m_Stack.PushOperandInt((int)c);
	}
	else //everything else is 4-byte
	{
		memcpy(&item, &pData[pField->iOffset], pField->iSize);
		m_Stack.PushOperand(item,pField->bReference);
	}
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_GoTo()
//-----------------------------------------------------------------------------
void cJVM::Inst_GoTo()
{
	short iOffset = (short)ReadU2();
	//offset is from beginnig of instruction, so -3
	m_pPC += iOffset - 3;
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_BALoad()
//-----------------------------------------------------------------------------
void cJVM::Inst_BALoad()
{
	//get array index
	int iIndex = m_Stack.PopOperand().iInt;
	//get object
	OBJECT *pObj = m_Stack.PopOperand().pObject;

	//check for null reference
	if(pObj == NULL)
	{
		Error("Error: Null reference.");
		return;
	}
	//check array index
	if(iIndex >= (int)pObj->iArraySize)
	{
		Error("Error: Array out of bounds.");
		return;
	}

	//skip object header
	u1 *pData = (u1*)(pObj + 1);

	//push array value
	m_Stack.PushOperandInt((int)pData[iIndex]);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_IInc()
//-----------------------------------------------------------------------------
void cJVM::Inst_IInc()
{
	//increament variable
	u1 iIndex = *m_pPC;
	++m_pPC;
	int iConst = *m_pPC;
	++m_pPC;

	m_Stack.IncVariableInt(iConst,iIndex);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_ArrayLength()
//-----------------------------------------------------------------------------
void cJVM::Inst_ArrayLength()
{
	//get array header
	OBJECT *pObj = m_Stack.PopOperand().pObject;

	//check for null reference
	if(pObj == NULL)
	{
		Error("Error: Null reference.");
		return;
	}

	//push array size
	m_Stack.PushOperandInt(pObj->iArraySize);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_If_ICMPLT()
//-----------------------------------------------------------------------------
void cJVM::Inst_If_ICMPLT()
{
	//get jump target (pc offset)
	short iOffset = (short)ReadU2();

	//pop values
	int iValue2 = m_Stack.PopOperand().iInt;
	int iValue1 = m_Stack.PopOperand().iInt;

	//if true
	if(iValue1 < iValue2)
	{
		//jump (offset from beginning of instruction -> -3)
		m_pPC += iOffset - 3;
	}
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_SIPush()
//-----------------------------------------------------------------------------
void cJVM::Inst_SIPush()
{
	//push short
	short s = (short)ReadU2();
	m_Stack.PushOperandInt(s);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_BAStore()
//-----------------------------------------------------------------------------
void cJVM::Inst_BAStore()
{
	//get value and array index
	int iValue = m_Stack.PopOperand().iInt;
	int iIndex = m_Stack.PopOperand().iInt;

	//get object
	OBJECT *pObj = m_Stack.PopOperand().pObject;

	//check for null reference
	if(pObj == NULL)
	{
		Error("Error: Null reference.");
		return;
	}
	//check array index
	if(iIndex >= (int)pObj->iArraySize)
	{
		Error("Error: Array out of bounds.");
		return;
	}

	//skip object's header
	u1 *pData = (u1*)(pObj + 1);

	//write byte
	pData[iIndex] = (char)iValue;
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_IALoad()
//-----------------------------------------------------------------------------
void cJVM::Inst_IALoad()
{
	//get inex and object
	int iIndex = m_Stack.PopOperand().iInt;
	OBJECT *pObj = m_Stack.PopOperand().pObject;

	//check for null reference
	if(pObj == NULL)
	{
		Error("Error: Null reference.");
		return;
	}
	//check array index
	if(iIndex >= (int)pObj->iArraySize)
	{
		Error("Error: Array out of bounds.");
		return;
	}

	//skip header
	u1 *pData = (u1*)(pObj + 1);

	int i;
	//copy value
	memcpy(&i,&pData[iIndex * sizeof(int)],sizeof(int));
	//push result
	m_Stack.PushOperandInt(i);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_AALoad()
//-----------------------------------------------------------------------------
void cJVM::Inst_AALoad()
{
	//get index and object
	int iIndex = m_Stack.PopOperand().iInt;
	OBJECT *pObj = m_Stack.PopOperand().pObject;

	//check for null reference
	if(pObj == NULL)
	{
		Error("Error: Null reference.");
		return;
	}
	//check array index
	if(iIndex >= (int)pObj->iArraySize)
	{
		Error("Error: Array out of bounds.");
		return;
	}

	//skip header
	OBJECT **ppData = (OBJECT**)(pObj + 1);

	//push result
	m_Stack.PushOperandObject(ppData[iIndex]);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_LDC2_W()
//-----------------------------------------------------------------------------
void cJVM::Inst_LDC2_W()
{
	//load double(long) from constant pool

	//constant pool index
	u2 iIndex = ReadU2();
	u4 iHighBytes, iLowBytes;
	
	//get value
	m_pCurClass->GetDoubleOrLongFromConstantPool(iIndex,iHighBytes,iLowBytes);
	//push both parts
	m_Stack.PushOperandInt(iHighBytes);
	m_Stack.PushOperandInt(iLowBytes);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_DAdd()
//-----------------------------------------------------------------------------
void cJVM::Inst_DAdd()
{
	double d1,d2;
	u4 iHighBytes,iLowBytes;
	//get first double
	iLowBytes = (u4)m_Stack.PopOperand().iInt;
	iHighBytes = (u4)m_Stack.PopOperand().iInt;

	//assembly first double
	this->GetDouble(d1,iHighBytes,iLowBytes);

	//same for second one
	iLowBytes = m_Stack.PopOperand().iInt;
	iHighBytes = m_Stack.PopOperand().iInt;

	this->GetDouble(d2,iHighBytes,iLowBytes);

	//add
	double d3 = d1 + d2;
	
	//disassembly result
	this->GetBytesFromDouble(d3,iHighBytes,iLowBytes);

	//push result
	m_Stack.PushOperandInt(iHighBytes);
	m_Stack.PushOperandInt(iLowBytes);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_DConst_0()
//-----------------------------------------------------------------------------
void cJVM::Inst_DConst_0()
{
	double d = 0;
	u4 iHighBytes,iLowBytes;
	//disassembly result
	this->GetBytesFromDouble(d,iHighBytes,iLowBytes);

	//push result
	m_Stack.PushOperandInt(iHighBytes);
	m_Stack.PushOperandInt(iLowBytes);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_DConst_1()
//-----------------------------------------------------------------------------
void cJVM::Inst_DConst_1()
{
	double d = 1;
	u4 iHighBytes,iLowBytes;
	//disassembly result
	this->GetBytesFromDouble(d,iHighBytes,iLowBytes);

	//push result
	m_Stack.PushOperandInt(iHighBytes);
	m_Stack.PushOperandInt(iLowBytes);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_D2I()
//-----------------------------------------------------------------------------
void cJVM::Inst_D2I()
{
	double d;
	u4 iHighBytes,iLowBytes;

	//get and assambly double
	iLowBytes = (u4)m_Stack.PopOperand().iInt;
	iHighBytes = (u4)m_Stack.PopOperand().iInt;

	this->GetDouble(d,iHighBytes,iLowBytes);

	//convert to integer
	int i = (int)d;
	//push result
	m_Stack.PushOperandInt(i);
}

//-----------------------------------------------------------------------------
// Name: cJVM::Inst_AAStore()
//-----------------------------------------------------------------------------
void cJVM::Inst_AAStore()
{
	OBJECT *pObj = m_Stack.PopOperand().pObject;
	int iIndex = m_Stack.PopOperand().iInt;

	OBJECT *pArrayObj = m_Stack.PopOperand().pObject;

	//check for null reference
	if(pArrayObj == NULL)
	{
		Error("Error: Null reference.");
		return;
	}
	//check array index
	if(iIndex >= (int)pArrayObj->iArraySize)
	{
		Error("Error: Array out of bounds.");
		return;
	}

	//skip header
	OBJECT **ppArray = (OBJECT**)(pArrayObj + 1);
	ppArray[iIndex] = pObj;
}

#define VM_DEBUG
//-----------------------------------------------------------------------------
// Name: cJVM::RunGC()
//-----------------------------------------------------------------------------
void cJVM::RunGC()
{
	#ifdef VM_DEBUG
	printf("\n\nVM: GC Started, allocated bytes: %d.\n",m_Heap.GetNumAllocatedBytes());
	#endif

	m_Heap.SwapHeapspace();

	//check stack
	for(u4 i = 2;i < this->m_Stack.GetTop();++i)
	{
		OBJECT *pObj = m_Stack.GetObject(i);
		if(pObj == NULL)
			continue;

		m_Stack.SetObject(i,CollectObject(pObj));
	}

	//check static members
	for(ClassMapItor i = m_Classes.begin();i != m_Classes.end();++i)
	{ //for all classes
		cJavaClass *pClass = i->second;
		//get static data pointer
		u1 *pData = pClass->m_pStaticData;

		//for all static fields
		for(u4 j = 0;j < pClass->GetNumFields();++j)
		{
			FIELD_INFO *pField = pClass->GetFieldInfo(j);
			if(pField->bReference == false)
				continue;
			if(pField->iAccessFlags & ACC_STATIC)
			{
				//get field
				OBJECT *pObj;
				memcpy(&pObj,&pData[pField->iOffset],sizeof(OBJECT*));
				if(pObj == NULL)
					continue;

				pObj = CollectObject(pObj);
				//collect field and update with new address
				memcpy(&pData[pField->iOffset],&pObj,sizeof(OBJECT*));
			}
		}
	}

	#ifdef VM_DEBUG
	printf("VM: GC Finished, allocated bytes: %d.\n\n",m_Heap.GetNumAllocatedBytes());
	#endif
}

//-----------------------------------------------------------------------------
// Name: cJVM::CollectObject()
//-----------------------------------------------------------------------------
OBJECT* cJVM::CollectObject(OBJECT *pObj)
{
	if(pObj->iArraySize == 0) //already handled, only update reference if neccesary
	{
		OBJECT *pNewObj = (OBJECT*)pObj->iSize;
		return (OBJECT*)pNewObj;
	}
	else
	{
		//check array size
		if(pObj->iArraySize > 1)
		{ 
			//array
			if(pObj->pClass != NULL)
			{ 
				//array of references -> check all
				//skip object header
				u1 *pObjData = (u1*)(pObj + 1);
				//go through all
				for(u4 i = 0;i < pObj->iArraySize;++i)
				{
					OBJECT *pArrayObj;
					memcpy(&pArrayObj,pObjData,sizeof(OBJECT*));

					if(pArrayObj == NULL)
					{
						pObjData += sizeof(OBJECT*);
						continue;
					}

					pArrayObj = CollectObject(pArrayObj);
					pObjData += sizeof(OBJECT*);
				}
			}

			//relocate array itself
			return m_Heap.RelocateObject(pObj);

		}
		else
		{ 
			//single obj - now for the difficult part
			u1 *pObjData = (u1*)(pObj + 1);

			cJavaClass *pClass = pObj->pClass;
			//go through all fields including super classes
			while(pClass != NULL)
			{
				for(int i = 0;i < pClass->GetNumFields();++i)
				{
					FIELD_INFO *pField = pClass->GetFieldInfo(i);
					if(pField->bReference == false)
						continue;
					if(pField->iAccessFlags & ACC_STATIC)
						continue;

					//get obj
					OBJECT *pFieldObj;
					memcpy(&pFieldObj,&pObjData[pField->iOffset],sizeof(OBJECT*));
					if(pFieldObj == NULL)
						continue;

					//collect obj
					memcpy(&pObjData[pField->iOffset],CollectObject(pFieldObj),sizeof(OBJECT*));
				}

				//try super class
				pClass = pClass->GetSuperClass();
			}			

			//relocate object itself
			return m_Heap.RelocateObject(pObj);
		}
	}
}