#include "string.h"
#include "stdio.h"
#include "Common.h"
#include "Heap.h"
#include "Class.h"
#include "Stack.h"

//-----------------------------------------------------------------------------
// Name: cStack::cStack()
//-----------------------------------------------------------------------------
cStack::cStack()
{
	Clear();
}

//-----------------------------------------------------------------------------
// Name: cStack::~cStack()
//-----------------------------------------------------------------------------
cStack::~cStack()
{
	m_Values.clear();
	m_Flags.clear();
}

//-----------------------------------------------------------------------------
// Name: cStack::cStack()
//-----------------------------------------------------------------------------
void cStack::Clear()
{
	m_iTopIndex = 0;
	m_iFrameIndex = 0;
	m_iVariablesIndex = 0;
	m_iOperandStackTopIndex = 0;
}

//-----------------------------------------------------------------------------
// Name: cStack::GetTop()
//-----------------------------------------------------------------------------
u4 cStack::GetTop() 
{ 
	return m_iTopIndex; 
}

//-----------------------------------------------------------------------------
// Name: cStack::Resize()
//-----------------------------------------------------------------------------
void cStack::Resize(u4 Size)
{
	m_Values.resize(Size);
	m_Flags.resize(Size,0);
	m_iTopIndex = 0;
	m_iFrameIndex = 0;
	m_iVariablesIndex = 0;
	m_iOperandStackTopIndex = 0;
}

//-----------------------------------------------------------------------------
// Name: cStack::PushMethodFrame()
//-----------------------------------------------------------------------------
bool cStack::PushMethodFrame(METHOD_INFO *pInfo,u1 *pPC,u2 iNumArgs)
{
	//compute frame's size
	u4 iSize = pInfo->iMaxLocals + pInfo->iMaxStack + 3;

	//copy arguments from operand stack	
	for(u2 i = 0;i < iNumArgs;++i)
	{
		--m_iOperandStackTopIndex;
		m_Values[m_iOperandStackTopIndex + 3] = m_Values[m_iOperandStackTopIndex];
		m_Flags[m_iOperandStackTopIndex + 3] = m_Flags[m_iOperandStackTopIndex];
	}

	m_iTopIndex = m_iOperandStackTopIndex;

	//store pc, current frame index and method info
	m_Values[m_iTopIndex].pPointer = pPC;
	m_Values[m_iTopIndex + 1].iInt = m_iFrameIndex;
	m_Values[m_iTopIndex + 2].pPointer = pInfo;

	m_Flags[m_iTopIndex] = 0;
	m_Flags[m_iTopIndex + 1] = 0;
	m_Flags[m_iTopIndex + 2] = 0;

	//update frame, variable and operand stack index
	m_iFrameIndex = m_iTopIndex;
	m_iTopIndex += iSize;
	m_iVariablesIndex = m_iFrameIndex + 3;
	m_iOperandStackTopIndex = m_iFrameIndex + 3 + pInfo->iMaxLocals;

	//zero frame
	memset(&m_Values[m_iFrameIndex + 3 + iNumArgs],0,iSize * sizeof(STACK_ITEM));
	memset(&m_Flags[m_iFrameIndex + 3 + iNumArgs],0,iSize * sizeof(u1));
	return true;
}

//-----------------------------------------------------------------------------
// Name: cStack::PopMethodFrame()
//-----------------------------------------------------------------------------
u1* cStack::PopMethodFrame(cJavaClass **ppCurClass)
{
	//retrieve program counter
	u1 *pCode = (u1*)m_Values[m_iFrameIndex].pPointer;
	
	//if pc == NULL, then this method is main
	//reset stack and terminate
	if(pCode == NULL)
	{
		m_iTopIndex = 0;
		m_iFrameIndex = 0;
		m_iVariablesIndex = 0;
		m_iOperandStackTopIndex = 0;
		*ppCurClass = NULL;
		return NULL;
	}

	//assign index before overwrite
	m_iOperandStackTopIndex = m_iFrameIndex;

	//METHOD_INFO *pInfo = (METHOD_INFO*)m_Values[m_iFrameIndex + 2].pPointer;
	m_iFrameIndex = m_Values[m_iFrameIndex + 1].iInt;

	//get method info
	METHOD_INFO *pInfo = (METHOD_INFO*)m_Values[m_iFrameIndex + 2].pPointer;

	//update stack
	m_iTopIndex = m_iFrameIndex + pInfo->iMaxLocals + pInfo->iMaxStack + 3;
	m_iVariablesIndex = m_iFrameIndex + 3;
	//m_iOperandStackTopIndex = m_iFrameIndex + 3 + pInfo->iMaxLocals;
	*ppCurClass = pInfo->pOwner;
	//return current pc
	return pCode;
}

//-----------------------------------------------------------------------------
// Name: cStack::GetVariable()
//-----------------------------------------------------------------------------
STACK_ITEM* cStack::GetVariable(u4 index)
{
	return &m_Values[m_iVariablesIndex + index];
}

//-----------------------------------------------------------------------------
// Name: cStack::SetVariableObject()
//-----------------------------------------------------------------------------
void cStack::SetVariableObject(OBJECT *pObj,u4 index)
{
	//set var and don't forget to raise reference flag
	m_Values[m_iVariablesIndex + index].pObject = pObj;
	m_Flags[m_iVariablesIndex + index] = 1;
}

//-----------------------------------------------------------------------------
// Name: cStack::SetVariableInt()
//-----------------------------------------------------------------------------
void cStack::SetVariableInt(int data,u4 index)
{
	m_Values[m_iVariablesIndex + index].iInt = data;
	m_Flags[m_iVariablesIndex + index] = 0; //not reference
}

//-----------------------------------------------------------------------------
// Name: cStack::IncVariableInt()
//-----------------------------------------------------------------------------
void cStack::IncVariableInt(int data,u4 index)
{
	m_Values[m_iVariablesIndex + index].iInt += data;
	m_Flags[m_iVariablesIndex + index] = 0; //not reference
}

//-----------------------------------------------------------------------------
// Name: cStack::PushVariable()
//-----------------------------------------------------------------------------
void cStack::PushVariable(u4 index)
{
	//take var and push it to the top of operand stack
	m_Values[m_iOperandStackTopIndex] = m_Values[m_iVariablesIndex + index];
	m_Flags[m_iOperandStackTopIndex] = m_Flags[m_iVariablesIndex + index];
	m_iOperandStackTopIndex++;
}

//-----------------------------------------------------------------------------
// Name: cStack::PushVariableDouble()
//-----------------------------------------------------------------------------
void cStack::PushVariableDouble(u4 index)
{
	//doubles take up to spaces
	m_Values[m_iOperandStackTopIndex] = m_Values[m_iVariablesIndex + index];
	m_Flags[m_iOperandStackTopIndex] = m_Flags[m_iVariablesIndex + index];
	m_iOperandStackTopIndex++;
	index++;
	m_Values[m_iOperandStackTopIndex] = m_Values[m_iVariablesIndex + index];
	m_Flags[m_iOperandStackTopIndex] = m_Flags[m_iVariablesIndex + index];
	m_iOperandStackTopIndex++;
}

//-----------------------------------------------------------------------------
// Name: cStack::PushOperandObject()
//-----------------------------------------------------------------------------
void cStack::PushOperandObject(OBJECT *pObj)
{
	m_Values[m_iOperandStackTopIndex].pObject = pObj;
	m_Flags[m_iOperandStackTopIndex] = 1; //reference
	++m_iOperandStackTopIndex;
}

//-----------------------------------------------------------------------------
// Name: cStack::PushOperand()
//-----------------------------------------------------------------------------
void cStack::PushOperand(STACK_ITEM Operand,bool bReference)
{
	m_Values[m_iOperandStackTopIndex] = Operand;
	m_Flags[m_iOperandStackTopIndex] = bReference;
	++m_iOperandStackTopIndex;
}

//-----------------------------------------------------------------------------
// Name: cStack::PushOperandInt()
//-----------------------------------------------------------------------------
void cStack::PushOperandInt(int i)
{
	m_Values[m_iOperandStackTopIndex].iInt = i;
	m_Flags[m_iOperandStackTopIndex] = 0; //not reference
	++m_iOperandStackTopIndex;
}

//-----------------------------------------------------------------------------
// Name: cStack::DuplicateTopOperand()
//-----------------------------------------------------------------------------
void cStack::DuplicateTopOperand()
{
	//duplicate single size operand
	u4 iIndex = m_iOperandStackTopIndex - 1;

	m_Values[m_iOperandStackTopIndex] = m_Values[iIndex];
	m_Flags[m_iOperandStackTopIndex] = m_Flags[iIndex];

	++m_iOperandStackTopIndex;
}

//-----------------------------------------------------------------------------
// Name: cStack::PopOperand()
//-----------------------------------------------------------------------------
STACK_ITEM cStack::PopOperand()
{
	return m_Values[--m_iOperandStackTopIndex];
}

//-----------------------------------------------------------------------------
// Name: cStack::GetOperand()
//-----------------------------------------------------------------------------
STACK_ITEM cStack::GetOperand(int index)
{
	//DEBUG PURPOSES
	return m_Values[index];
}

//-----------------------------------------------------------------------------
// Name: cStack::GetObject()
//-----------------------------------------------------------------------------
OBJECT* cStack::GetObject(int index)
{
	if(m_Flags[index] == 1)
		return m_Values[index].pObject;
	return NULL;
}

//-----------------------------------------------------------------------------
// Name: cStack::SetObject()
//-----------------------------------------------------------------------------
void cStack::SetObject(int index,OBJECT* pObj)
{
	m_Flags[index] = 1;
	m_Values[index].pObject = pObj;
}