#include <vector>

//-----------------------------------------------------------------------------
//  Name: struct STACK_ITEM
//! Represents heap allocated object header
//-----------------------------------------------------------------------------
typedef struct STACK_ITEM
{
	union
	{
		OBJECT	*pObject;
		int		iInt;
		float	fFloat;
		void	*pPointer;
	};

} STACK_ITEM;

//-----------------------------------------------------------------------------
// Name: class cStack
// Desc: Implements script's stack
//-----------------------------------------------------------------------------
class cStack
{
private:
	typedef std::vector<STACK_ITEM> StackVec;
	typedef std::vector<u1> FlagVec;

	//Vector of stack values
	StackVec	m_Values;
	//Stack values flags
	FlagVec		m_Flags;
	//Index to top of the stack
	u4		m_iTopIndex;
	//Top Frame index
	u4		m_iFrameIndex;
	//Frame's variables index
	u4		m_iVariablesIndex;
	//Frame's operand stack top index
	u4		m_iOperandStackTopIndex;

public:
	//Constructor
	cStack();
	//Destructor
	~cStack();

	//init(clear) stack
	void Clear();

	//Get top of the stack
	u4 GetTop();
	//Resize stack
	void Resize(u4 Size);
	//Push method's frame on top of stack
	bool PushMethodFrame(METHOD_INFO *pInfo,u1 *pPC,u2 iNumArgs);
	//Pop method's frame from top of stack
	u1* PopMethodFrame(cJavaClass **ppCurClass);
	//Return current frame's variable with index
	STACK_ITEM* GetVariable(u4 index);
	//Set current frame's variable - object
	void SetVariableObject(OBJECT *pObj,u4 index);
	//Set current frame's variable - int
	void SetVariableInt(int data,u4 index);
	//Increment current frame's variable - int
	void IncVariableInt(int data,u4 index);
	//Get variable and push it on top of operand stack
	void PushVariable(u4 index);
	//Get variable and push it on top of operand stack - two values (double, long)
	void PushVariableDouble(u4 index);
	//Push object on top of operand stack
	void PushOperandObject(OBJECT *pObj);
	//Push operand on top of operand stack
	void PushOperand(STACK_ITEM Operand,bool bReference);
	//Push int on top of operand stack
	void PushOperandInt(int i);
	//Duplicate top operand on operand stack
	void DuplicateTopOperand();
	//Pop operand from top of stack
	STACK_ITEM PopOperand();
	//Get operand with specified index
	STACK_ITEM GetOperand(int index);


	//Used during GC. Checks if item at index is an object reference, if so,
	//pointer is returned. Otherwise NULL is returned
	OBJECT* GetObject(int index);
	//Used during GC. Sets object reference to specified stack item.
	void SetObject(int index,OBJECT *pObj);
};